/*
 * XML By Example, chapter 8: SAX
 */

package com.psol.xbe;

import java.util.*;
import org.xml.sax.*;
import org.xml.sax.helpers.ParserFactory;

/**
 * Starting point class: initializes the parser, creates the
 * various objects, etc.
 * @author bmarchal@pineapplesoft.com
 */

public class BestDeal
{
   /**
    * the parser class (IBM's XML for Java)
    */
   private static final String 
      PARSER_NAME = "com.ibm.xml.parsers.SAXParser";

   /**
    * main() method
    * decodes command-line parameters and invokes the parser
    * @param args command-line argument
    * @throw Exception catch-all for underlying exceptions
    */
   public static void main(String[] args)
      throws Exception
   {
      if(args.length < 2)
      {
         System.out.println("java com.psol.xbe.BestDeal filename delivery");
         return;
      }

      ComparingMachine comparingMachine =
         new ComparingMachine(Integer.parseInt(args[1]));
      SAX2Internal sax2Internal =
         new SAX2Internal(comparingMachine);

      try
      {
         Parser parser = ParserFactory.makeParser(PARSER_NAME);
         parser.setDocumentHandler(sax2Internal);
         parser.parse(args[0]);
      }
      catch(SAXException e)
      {
         Exception x = e.getException();
         if(null != x)
            throw x;
         else
            throw e;
      }

      System.out.println("The best deal is proposed by " +
                         comparingMachine.getVendor());
      System.out.println("a " +
                         comparingMachine.getProductName() +
                         " at " + comparingMachine.getPrice() + 
                         " delivered in " + 
                         comparingMachine.getDelivery() +
                         " days");
   }
}

/**
 * This class receives events from the SAX2Internal adapter
 * and does the comparison required.
 * This class holds the "business logic."
 */

class ComparingMachine
{
   /**
    * properties we are collecting: best price
    */
   protected double bestPrice = Double.MAX_VALUE;
   
   /**
    * properties we are collecting: delivery time
    */
   protected int proposedDelivery = Integer.MAX_VALUE;

   /**
    * properties we are collecting: product and vendor names
    */
   protected String productName = null,
                    vendorName = null;

   /**
    * target delivery value (we refuse elements above this target)
    */
   protected int targetDelivery;

   /**
    * creates a ComparingMachine
    * @param td the target for delivery
    */
   public ComparingMachine(int td)
   {
      targetDelivery = td;
   }

   /**
    * called by SAX2Internal when it has found the product name
    * @param name the product name
    */
   public void setProductName(String name)
   {
      productName = name;
   }

   /**
    * called by SAX2Internal when it has found a price
    * @param vendor vendor's name
    * @param price price proposal
    * @param delivery delivery time proposal
    */
   public void compare(String vendor,double price,int delivery)
   {
      if(delivery <= targetDelivery)
      {
         if(bestPrice > price)
         {
            bestPrice = price;
            vendorName = vendor;
            proposedDelivery = delivery;
         }
      }
   }

   /**
    * property accessor: vendor's name
    * @return the vendor with the cheapest offer so far
    */
   public String getVendor()
   {
      return vendorName;
   }

   /**
    * property accessor: best price
    * @return the best price so far
    */
   public double getPrice()
   {
      return bestPrice;
   }

   /**
    * property accessor: proposed delivery
    * @return the proposed delivery time
    */
   public int getDelivery()
   {
      return proposedDelivery;
   }

   /**
    * property accessor: product name
    * @return the product name
    */
   public String getProductName()
   {
      return productName;
   }
}

/**
 * SAX event handler to adapt from the SAX interface to
 * whatever the application uses internally.
 */
class SAX2Internal
   extends HandlerBase
{
   /**
    * state constants
    */
   final protected int START = 0,
                       PRODUCT = 1,
                       PRODUCT_NAME = 2,
                       VENDOR = 3,
                       VENDOR_NAME = 4,
                       VENDOR_PRICE = 5;

   /**
    * the current state
    */
   protected int state = START;

   /**
    * current leaf element and current vendor
    */
   protected LeafElement currentElement = null,
                         currentVendor = null;

   /**
    * BestDeal object this event handler interfaces with
    */
   protected ComparingMachine comparingMachine;

   /**
    * creates a SAX2Internal
    * @param cm the ComparingMachine to interface with
    */
   public SAX2Internal(ComparingMachine cm)
   {
      comparingMachine = cm;
   }

   /**
    * startElement event
    * @param name element's name
    * @param attributes element's attributes
    */
   public void startElement(String name,AttributeList attributes)
   {
      // this accepts many combinations of elements
      // it would work if new elements where being added, etc.
      // this ensures maximal flexibility: if the document
      // has to be validated, a validating parser does it
      switch(state)
      {
         case START:
            if(name.equals("product"))
               state = PRODUCT;
            break;
         case PRODUCT:
            if(name.equals("name"))
            {
               state = PRODUCT_NAME;
               currentElement = new LeafElement(name,attributes);
            }
            if(name.equals("vendor"))
               state = VENDOR;
            break;
         case VENDOR:
            if(name.equals("name"))
            {
               state = VENDOR_NAME;
               currentElement = new LeafElement(name,attributes);
            }
            if(name.equals("price"))
            {
               state = VENDOR_PRICE;
               currentElement = new LeafElement(name,attributes);
            }
            break;
      }
   }

   /**
    * content of the element
    * @param chars documents characters
    * @param start first character in the content
    * @param length last character in the content
    */
   public void characters(char[] chars,int start,int length)
   {
      switch(state)
      {
         case PRODUCT_NAME:
         case VENDOR_NAME:
         case VENDOR_PRICE:
            currentElement.append(chars,start,length);
            break;
      }
   }

   /**
    * endElement event
    * @param name element's name
    */
   public void endElement(String name)
   {
      switch(state)
      {
         case PRODUCT_NAME:
            if(name.equals("name"))
            {
               state = PRODUCT;
               comparingMachine.setProductName(
                  currentElement.getText());
            }
            break;
         case VENDOR:
            if(name.equals("vendor"))
               state = PRODUCT;
            break;
         case VENDOR_NAME:
            if(name.equals("name"))
            {
               state = VENDOR;
               currentVendor = currentElement;
            }
            break;
         case VENDOR_PRICE:
            if(name.equals("price"))
            {
               state = VENDOR;
               double price = toDouble(currentElement.getText());
               Dictionary attributes =
                  currentElement.getAttributes();
               String stringDelivery =
                  (String)attributes.get("delivery");
               int delivery = Integer.parseInt(stringDelivery);
               comparingMachine.compare(currentVendor.getText(),
                                        price,
                                        delivery);
            }
            break;
      }
   }

   /**
    * helper method: turn a string in a double
    * @param string number as a string
    * @return the number as a double, or 0.0 if it cannot convert
    * the number
    */
   protected double toDouble(String string)
   {
      Double stringDouble = Double.valueOf(string);
      if(null != stringDouble)
         return stringDouble.doubleValue();
      else
         return 0.0;
   }
}

/*
 * helper class: used to store a leaf element content
 */
class LeafElement
{
   /**
    * property: element's name
    */
   protected String name;

   /**
    * property: element's attributes
    */
   protected Dictionary attributes;

   /**
    * property: element's text
    */
   protected StringBuffer text = new StringBuffer();

   /**
    * creates a new element
    * @param n element's name
    * @param a element's attributes
    */
   public LeafElement(String n,AttributeList al)
   {
      name = n;
      attributes = new Hashtable();
      for(int i = 0;i < al.getLength();i++)
         attributes.put(al.getName(i),al.getValue(i));
   }

   /**
    * append to the current text
    * @param chars array of characters
    * @param start where to start in chars
    * @param length how many characters to consider in chars
    */
   public void append(char[] chars,int start,int length)
   {
      text.append(chars,start,length);
   }

   /**
    * property accessor: text
    */
   public String getText()
   {
      return text.toString();
   }

   /**
    * property accessor: name
    */
   public String getName()
   {
      return name;
   }

   /**
    * property accessor: attributes
    */
   public Dictionary getAttributes()
   {
      return attributes;
   }
}